@@ -1252,3 +1252,64 @@ fn check_debugger() {
     let result = verifier.verify_without_limit(script_version, &rtx);
     assert!(result.is_ok(), "result {:?}", result);
 }
+
+#[test]
+fn check_typical_secp256k1_blake160_2_in_2_out_resume_load_cycles() {
+    _check_typical_secp256k1_blake160_2_in_2_out_resume_load_cycles(23);
+    _check_typical_secp256k1_blake160_2_in_2_out_resume_load_cycles(34);
+    _check_typical_secp256k1_blake160_2_in_2_out_resume_load_cycles(44);
+}
+
+fn _check_typical_secp256k1_blake160_2_in_2_out_resume_load_cycles(step_cycles: Cycle) {
+    const LOAD_CYCLES: Cycle = 25356;
+
+    let script_version = SCRIPT_VERSION;
+    let rtx = random_2_in_2_out_rtx();
+    let mut cycles = 0;
+    let verifier = TransactionScriptsVerifierWithEnv::new();
+
+    let result = verifier.verify_map(script_version, &rtx, |verifier| {
+        let mut init_state: Option<TransactionState<'_>> = None;
+
+        if let VerifyResult::Suspended(state) = verifier.resumable_verify(step_cycles).unwrap() {
+            init_state = Some(state);
+        }
+
+        loop {
+            let state = init_state.take().unwrap();
+            let (limit_cycles, _last) = state.next_limit_cycles(step_cycles, TWO_IN_TWO_OUT_CYCLES);
+            match verifier.resume_from_state(state, limit_cycles).unwrap() {
+                VerifyResult::Suspended(state) => init_state = Some(state),
+                VerifyResult::Completed(cycle) => {
+                    cycles = cycle;
+                    break;
+                }
+            }
+        }
+
+        verifier.verify(TWO_IN_TWO_OUT_CYCLES)
+    });
+
+    let cycles_once = result.unwrap();
+    assert!(
+        cycles <= TWO_IN_TWO_OUT_CYCLES,
+        "step_cycles {}",
+        step_cycles
+    );
+    assert!(
+        cycles >= TWO_IN_TWO_OUT_CYCLES - CYCLE_BOUND,
+        "step_cycles {}",
+        step_cycles
+    );
+
+    if SCRIPT_VERSION >= ScriptVersion::V1 {
+        assert_eq!(cycles, cycles_once, "step_cycles {}", step_cycles);
+    } else {
+        assert_eq!(
+            cycles + LOAD_CYCLES,
+            cycles_once,
+            "step_cycles {}",
+            step_cycles
+        );
+    }
+}
